/**********************************************************************
 *
 *   FreeDoko a Doppelkopf-Game
 *    
 *   Copyright (C) 2001-2004  by Diether Knof and Borg Enders
 *
 *   This program is free software; you can redistribute it and/or 
 *   modify it under the terms of the GNU General Public License as 
 *   published by the Free Software Foundation; either version 2 of 
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details. 
 *   You can find this license in the file 'gpl.txt'.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *   MA  02111-1307  USA
 *
 *  Contact:
 *    Diether Knof dknof@gmx.de
 *    Borg Enders  borg@borgsoft.de
 *
 **********************************************************************/

#ifndef HEADER_CONSTANTS
#define HEADER_CONSTANTS

// using the microsoft visual C++ compiler
#ifdef _MSC_VER
#define MS_VISUAL_C
#define WINDOWS
#endif

// C++ include

#include <typeinfo>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <algorithm>

using std::cout;
using std::cerr;
using std::endl;

// for temporary output
#define COUT cout

#include <vector>
#include <list>
#include <map>
#include <set>
#include <string>



#ifdef WINDOWS
#define USE_REGISTRY
#endif

#ifdef WINDOWS
#ifdef USE_REGISTRY
#include <windows.h>
#undef min
#undef max
#endif
#endif


#ifndef MS_VISUAL_C
#include <dirent.h>
#endif


#include "utils/version.h"


// Whether GTKMM can be used.
// When not defined, FreeDoko should compile and run without any GTKMM library
// this makro shall be defined through the compiler (in the makefile)
//#define GTKMM


// code, that is postphoned -- leave this macro undefined
#undef POSTPHONED

// code that is workaround -- leave this macro defined
#define WORKAROUND


extern DK::Utils::Version const version;
extern std::vector<DK::Utils::Version const*> all_versions;
#define CONTACT "freedoko@users.sourceforge.net"


// In the release-version there are some elements hidden
//#define RELEASE

#ifdef RELEASE
#define NO_CRITICAL_CHECKS
#endif

// All players are AI, there is no pause and no 'full trick'-window
extern bool FAST_PLAY;

// mutate the rules after each game
extern bool MUTATE_RULE;

// There is only output of the game seed
extern bool const SEED_OUT;
// the starting seed
extern unsigned SEED_START;
// the starting player
extern unsigned const STARTPLAYER;

#include "info.h"

// some utils (RAND(int), MARK(), expand_filename(string), Null_Str
#include "utils.h"


// whether an assertion shall create a segmentation fault (for a core dump)
// (define it in 'Makefile.local')
//#define ASSERTION_GENERATES_SEGFAULT

#ifdef RELEASE
// if DEBUG_NO_OUTPUT is defined, debugging information are not written
#define DEBUG_NO_OUTPUT
#endif

// if DEBUG_NO is defined, the internal checks are skipped
// this speeds up the program up to factor 4
//#define DEBUG_NO

// debugging utils
#include "debug.h"

#define SEGFAULT (*reinterpret_cast<int*>(NULL) = 0)

#endif // #ifndef HEADER_CONSTANTS
